# -*- coding: utf-8 -*-
"""
Logic specific to trial databases.
"""
import binascii
import contextlib
import errno
import logging
import os
import random
import shutil
import subprocess
import time

from datetime import datetime
from dateutil.relativedelta import relativedelta

import requests
import werkzeug.urls
from werkzeug.exceptions import NotFound

try:
    import metabase
    import pg_util
    from auth_util import SaasAuth
    from const import EDU_PREFIX, OAUTH_BASE_URL, SPECIAL_PREFIXES, TMPFROZENDIR, SAAS_BASE_URL, TEST, TIMEOUT
    from plausible_api import PlausibleAPI
except ImportError:
    from . import metabase
    from . import pg_util
    from .auth_util import SaasAuth
    from .const import EDU_PREFIX, OAUTH_BASE_URL, SPECIAL_PREFIXES, TMPFROZENDIR, SAAS_BASE_URL, TEST, TIMEOUT
    from .plausible_api import PlausibleAPI


_logger = logging.getLogger(__name__)


FILESTOREDIR = os.path.expanduser("~/filestore")


def bump_sliding_expiration(mdb):
    """
    If the DB has a sliding expiration (one app free), increase it.

    Invoked at session load in the saas_trial module.
    """

    def get_increase(lifetime, long_lifetime_inc):
        if lifetime > 180:
            return long_lifetime_inc
        if lifetime > 90:
            return relativedelta(months=6)
        if lifetime > 7:
            return relativedelta(months=3)
        return relativedelta(days=15)

    lifetime = (datetime.utcnow() - mdb.create_date).days
    delta = None

    if (
        mdb.mode == "trial"
        and not mdb.extra_apps
        and mdb.status == "activated"
        and not mdb.name.startswith(SPECIAL_PREFIXES)
    ):
        delta = get_increase(lifetime, relativedelta(years=1))

    if (
        mdb.mode == "trial"
        and mdb.status == "activated"
        and mdb.name.startswith(EDU_PREFIX)
        and lifetime < 730  # Max two years
    ):
        max_expiry = min(730 - lifetime, 365)
        delta = get_increase(lifetime, relativedelta(days=max_expiry))

    if delta:
        # metabase.update has probably just been called, and it changes mdb.last_cnx_date in its own cursor
        last_cnx_date = metabase.get(mdb.name).last_cnx_date or datetime.utcnow()
        new_expire_date = last_cnx_date + delta
        # Sliding expiration should only extend the validity
        if new_expire_date > mdb.expire_date:
            metabase.update(mdb, expire_date=new_expire_date.isoformat(' '))

def create_plausible_site(dbname, tz=None):
    MAX_TRY = 3
    res = {}
    plausible = PlausibleAPI()
    domain = '%s.%s' % (dbname, SAAS_BASE_URL)
    _try = 0
    # while != {'error': 'domain has already been taken'}
    while (plausible.create_website(domain, tz=tz).get('error') and _try < MAX_TRY):
        _try += 1
        rand = "%04x" % random.getrandbits(16)
        domain = '%s.%s-%s' % (dbname, SAAS_BASE_URL, rand)

    if _try < MAX_TRY:
        plausible.create_goal(domain, 'event', 'Lead Generation')
        plausible.create_goal(domain, 'event', 'Shop')
        plausible.create_goal(domain, 'event', '404')
        plausible.create_goal(domain, 'event', 'Custom')
        plausible.ping(domain)  # Need at least one event in clickhouse to be recycled after 15 days inactivity
        resp = plausible.create_sharedlinks(domain, "Hosted by Odoo")
        res['auth'] = werkzeug.urls.url_decode(werkzeug.urls.url_parse(resp.get('url', '')).query).get('auth')
        res['domain'] = domain
    return res


class Frozen:
    """
    Logic for handling frozen trials (4h trial that did not get activated).
    This class does not directly manipulate the metabase since that will be the
    responsibility of the caller.

     - after 4h, the database is archived then dropped and the status is changed to
       "frozen".  The user is then redirected to a page prompting them to activate
     - if they activate, the database is "thawed"
     - if they still don't activate within 2 weeks, we remove the archive and
       permanently remove the DB from the metabase.

    Note, the class is used only for namespacing and has only static methods.
    """

    @staticmethod
    def get_frozen_status(db_info):
        """
        Return true if the db is frozen and the thaw files are present.
        """
        dump_fn = os.path.join(TMPFROZENDIR, db_info.name + ".dmp")
        if db_info.status == "frozen" and not os.path.exists(dump_fn):
            # will happen if the dump files were lost (e.g., the server was moved)
            raise NotFound()
        return db_info.status == "frozen"

    @staticmethod
    def freeze_trial(db):
        """
        Archive the DB and its filestore to a non-backed up area
        (TMPFROZENDIR).
        """
        dest_fn = os.path.join(TMPFROZENDIR, db.name + ".dmp")
        pg_util.dump(db.name, dest_fn)
        fs = os.path.join(FILESTOREDIR, db.name)
        dest_fn = os.path.join(TMPFROZENDIR, db.name + ".filestore.tar.gz")
        cmd = ["tar", "-czf", dest_fn, "."]
        subprocess.call(cmd, cwd=fs)
        _logger.info("frozen trial db %s", db.name)

    @staticmethod
    def drop_archive(db):
        """
        Remove the generated archives for the db, logging (but not propagating) errors.
        """
        with Frozen._archive_lock(db, "dropping") as lck:
            if not lck:
                return
            for suffix in [".dmp", ".filestore.tar.gz"]:
                fn = os.path.join(TMPFROZENDIR, db.name + suffix)
                try:
                    if os.path.exists(fn):
                        os.unlink(fn)
                except Exception:
                    _logger.error(
                        "%s: could not remove db thaw file %s", db.name, fn, exc_info=True
                    )

    @staticmethod
    def thaw_trial(db):
        """If the db is a frozen trial, initiate a thaw (if not already in process)
        and return true.
        If a thaw is already in process, return False.
        Otherwise, return false.
        """
        dump_fn = os.path.join(TMPFROZENDIR, db.name + ".dmp")
        fs_fn = os.path.join(TMPFROZENDIR, db.name + ".filestore.tar.gz")
        if not (os.path.exists(dump_fn) and os.path.exists(fs_fn)):
            # not common, because /frozen/status should have given an error when first called
            _logger.warning(
                "%s: database thaw files (%s, %s) are not present", db.name, dump_fn, fs_fn
            )
            return False
        with Frozen._archive_lock(db, "thawing") as lck:
            if not lck:
                return False
            try:
                pg_util.restore(db.name, dump_fn)
                fs = os.path.join(FILESTOREDIR, db.name)
                try:
                    os.mkdir(fs)
                except OSError as e:
                    if e.errno != errno.EEXIST:
                        raise
                cmd = ["tar", "-xzf", fs_fn]
                subprocess.call(cmd, cwd=fs)
                os.unlink(dump_fn)
                os.unlink(fs_fn)
            except Exception:
                pg_util.drop(db.name)
                raise
        metabase.reset_cron_round(db.name)
        return True

    @staticmethod
    def should_freeze(db_info):
        """
        Return true if the db should be frozen.
        This is called from trials.gc for trials that are in the to_activate status,
        so we already know that it has reached expiration date.
        """
        return (
            # we only freeze dbs on the "new" servers, because otherwise we might pick up duplicates
            # where the source_uuid was incorrectly reset to null - for example, in the move db process
            metabase.has_role("new")
            # do not include duplicates
            and db_info.source_uuid is None
        )

    @staticmethod
    def _possibly_frozen(database):
        """
            database is a recordset from model openerp.enterprise.database
            return True if the database is possibly frozen, which means
            they are not activated, not duplicated or not out of the trial period
        """
        return (
            database.confirm_token and not database.parent_id and
            not database.subscription_id and
            datetime.utcnow() > database.create_date + relativedelta(hours=4) and
            datetime.utcnow() < database.create_date + relativedelta(weeks=2, hours=4)
        )

    @staticmethod
    @contextlib.contextmanager
    def _archive_lock(db, reason):
        """Prevent concurrent access to a frozen db restore files.
        If a lock cannot be obtained, it will log a warning using the provided
        reason and return False."""
        # so we use a file lock, rather than something smarter like a "SELECT FOR UPDATE", to avoid
        # locking the DB (bad enough that we are hogging a web worker for this...)
        # ideally this would execute in an async job queue that could make sure we don't requeue
        # the job while it's in progress, and free up the web worker
        lock_fn = os.path.join(TMPFROZENDIR, db.name + ".lock")
        try:
            if time.time() - os.path.getmtime(lock_fn) > 15 * 60:
                # after a few min, remove a stale lock, otherwise we'll never get rid of the archive
                # these dbs should normally restore much faster than this since they have only
                # been used 4h
                _logger.warning("%s: remove stale lock file at %s" % (db.name, lock_fn))
                os.unlink(lock_fn)
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise
        try:
            os.close(os.open(lock_fn, os.O_EXCL | os.O_CREAT))
            yield True
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
            _logger.warning(
                "%s: could not lock for %s, database thaw is in progress", db.name, reason
            )
            yield False
        finally:
            os.unlink(lock_fn)


def drop(dbname):
    dbname = dbname.lower()
    _logger.info("Drop database %s", dbname)
    pg_util.drop(dbname)
    fs = os.path.join(FILESTOREDIR, dbname)
    shutil.rmtree(fs, ignore_errors=True)
    _logger.info("DROP DB FILESTORE: %s", fs)


def gc():
    """
    Drop / freeze expired trial databases.
    """
    for db in metabase.get_expired_dbs():
        drop_db = True
        set_expiry, set_mode, set_status = None, None, None
        try:
            if db.mode == "trial":
                if db.status == "to_activate":
                    # 4h trials that have not been activated
                    if Frozen.should_freeze(db):
                        set_expiry = "2 weeks"
                        set_status = "frozen"
                        # 'Best effort': freeze may fail for several reasons (e.g. the DB
                        # doesn't exist in postgres).
                        try:
                            Frozen.freeze_trial(db)
                        except:
                            _logger.warning("Could not freeze DB %s", db.name)
                            set_expiry = "1 hour"
                            set_mode = "drop"
                            set_status = "unsubscribed"
                    else:
                        # duplicate, etc: drop them
                        pass
                elif db.status == "frozen":
                    # time to go
                    Frozen.drop_archive(db)
                    # since it was dropped before no need to do it now
                    drop_db = False
                elif db.status == "activated":
                    # Activated databases are dropped later unless they are duplicated, in
                    # which case they are dropped now.
                    if not db.source_uuid:
                        drop_db = False
                        set_expiry = "3 weeks"
                        set_mode = "drop"
                        set_status = "unsubscribed"
                else:
                    # not a normal combination
                    # log it just in case, but let it go through the default handling and get dropped
                    _logger.warning("Unknown status for trial in gc: %s" % (db.status,))
            if drop_db:
                drop(db.name)
            if set_expiry:
                metabase.set_expiry_in(db.name, set_expiry, set_mode, set_status)
            else:
                metabase.remove_database(db.name)
        except Exception as x:
            _logger.error("Cannot drop database %s: skipping" % (db.name,), exc_info=x)

    # Clean-up ghost databases
    for db_name in metabase.get_ghost_db_names():
        _logger.info("Drop ghost database %s", db_name)
        metabase.set_expiry_in(db_name, "0 day", "drop", "unsubscribed")

    # Clean-up failed db creation
    for db_name in metabase.get_failed_trial_creation():
        _logger.info("Drop failed database creation %s", db_name)
        metabase.set_expiry_in(db_name, "0 day", "drop", "unsubscribed")


def create_spare_database():
    """
    Provision a few spare databases.
    """
    MIN_SPARE_DB = 10
    SPARE_PREFIX = "__spare_"

    while len(metabase.spare_list()) < MIN_SPARE_DB:
        uid = binascii.hexlify(os.urandom(4)).decode()
        dbname = SPARE_PREFIX + uid
        pg_util.createdb(dbname)
        metabase.add(dbname, mode="spare")
